# lingo - payment and billing system
# Copyright (C) 2022-2024  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import difflib
import json

from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.formats import date_format
from django.utils.timezone import localtime
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, ListView


class InstanceWithSnapshotHistoryView(ListView):
    def get_queryset(self):
        self.instance = get_object_or_404(self.model.get_instance_model(), pk=self.kwargs['pk'])
        return self.instance.instance_snapshots.all().select_related('user')

    def get_context_data(self, **kwargs):
        kwargs[self.instance_context_key] = self.instance
        kwargs['object'] = self.instance
        current_date = None
        context = super().get_context_data(**kwargs)
        day_snapshot = None
        for snapshot in context['object_list']:
            if snapshot.timestamp.date() != current_date:
                current_date = snapshot.timestamp.date()
                snapshot.new_day = True
                snapshot.day_other_count = 0
                day_snapshot = snapshot
            else:
                day_snapshot.day_other_count += 1
        return context


class InstanceWithSnapshotHistoryCompareView(DetailView):
    def get_snapshots_from_application(self):
        version1 = self.request.GET.get('version1')
        version2 = self.request.GET.get('version2')
        if not version1 or not version2:
            raise Http404

        snapshot_for_app1 = (
            self.model.get_snapshot_model()
            .objects.filter(
                instance=self.object,
                application_slug=self.request.GET['application'],
                application_version=self.request.GET['version1'],
            )
            .order_by('timestamp')
            .last()
        )
        snapshot_for_app2 = (
            self.model.get_snapshot_model()
            .objects.filter(
                instance=self.object,
                application_slug=self.request.GET['application'],
                application_version=self.request.GET['version2'],
            )
            .order_by('timestamp')
            .last()
        )
        return snapshot_for_app1, snapshot_for_app2

    def get_snapshots(self):
        if 'application' in self.request.GET:
            return self.get_snapshots_from_application()

        id1 = self.request.GET.get('version1')
        id2 = self.request.GET.get('version2')
        if not id1 or not id2:
            raise Http404

        snapshot1 = get_object_or_404(self.model.get_snapshot_model(), pk=id1, instance=self.object)
        snapshot2 = get_object_or_404(self.model.get_snapshot_model(), pk=id2, instance=self.object)

        return snapshot1, snapshot2

    def get_context_data(self, **kwargs):
        kwargs[self.instance_context_key] = self.object

        mode = self.request.GET.get('mode') or 'json'
        if mode not in ['json']:
            raise Http404

        snapshot1, snapshot2 = self.get_snapshots()
        if not snapshot1 or not snapshot2:
            return redirect(reverse(self.history_view, args=[self.object.pk]))
        if snapshot1.timestamp > snapshot2.timestamp:
            snapshot1, snapshot2 = snapshot2, snapshot1

        kwargs['mode'] = mode
        kwargs['snapshot1'] = snapshot1
        kwargs['snapshot2'] = snapshot2
        kwargs['fromdesc'] = self.get_snapshot_desc(snapshot1)
        kwargs['todesc'] = self.get_snapshot_desc(snapshot2)
        kwargs.update(getattr(self, 'get_compare_%s_context' % mode)(snapshot1, snapshot2))

        return super().get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)
        if isinstance(context, HttpResponseRedirect):
            return context
        return self.render_to_response(context)

    def get_compare_json_context(self, snapshot1, snapshot2):
        s1 = json.dumps(snapshot1.serialization, sort_keys=True, indent=2)
        s2 = json.dumps(snapshot2.serialization, sort_keys=True, indent=2)
        diff_serialization = difflib.HtmlDiff(wrapcolumn=160).make_table(
            fromlines=s1.splitlines(True),
            tolines=s2.splitlines(True),
        )

        return {
            'diff_serialization': diff_serialization,
        }

    def get_snapshot_desc(self, snapshot):
        label_or_comment = ''
        if snapshot.label:
            label_or_comment = snapshot.label
        elif snapshot.comment:
            label_or_comment = snapshot.comment
        if snapshot.application_version:
            label_or_comment += ' (%s)' % _('Version %s') % snapshot.application_version
        return '{name} ({pk}) - {label_or_comment} ({user}{timestamp})'.format(
            name=_('Snapshot'),
            pk=snapshot.id,
            label_or_comment=label_or_comment,
            user='%s ' % snapshot.user if snapshot.user_id else '',
            timestamp=date_format(localtime(snapshot.timestamp), format='DATETIME_FORMAT'),
        )
